'''Tools to graph revisions in revision control systems.

Currently only tla <http://gnuarch.org/> archives can be graphed, and output is
for dot, part of graphviz <http://www.research.att.com/sw/tools/graphviz/>.
However, the design allows for more input or output modules to be implemented,
so this will likely change some day.

Written by Phil Frost <indigo@bitglue.com>. Public Domain. If you like this
software please consider giving money to Unununium <http://unununium.org/>.
'''

from sets import Set, ImmutableSet
from itertools import chain, ifilter

from pprint import pprint
import sys



class Node( str ):
    '''A node representing a revision in the Graph.

    Nodes a strings which are unique identifiers for them, with associated
    metadata.

    attributes:
    -----------
    - parent: The direct parent of this node, or None
    - merged_nodes, unmerged_nodes: an iterable container of all nodes that
      were merged or unmerged into this node
    - label: a short string label for this node
    - group: a string that groups related nodes, such as by branch
    - start_date, end_date: The start and end commit dates as UTC datetime
      objects of this node. Only start_date should be set by the input module;
      end_date is set iff the node is folded.
    - summary: a short string describing the commit
    - trunk: True if this is a trunk node, that is, directly requested by the
      user or a direct decendant.
    '''
    parent = None
    merged_nodes = []
    unmerged_nodes = []
    label = None
    group = None
    start_date = end_date = None
    summary = ''
    trunk = False


    def all_referenced_nodes( self ):
        '''Return iterable for all nodes referenced by this one.

        May contain duplicates.
        '''
        return ifilter( lambda x: x is not None, chain(
        [self.parent],
        self.merged_nodes,
        self.unmerged_nodes ) )


#    def __repr__( self ):
#        return '<%s.%s object at 0x%x>' % (self.__module__, self.__class__.__name__, id(self))


    def __new__( klass, nodename, *args, **kwargs ):
        return str.__new__( klass, nodename )


    def is_adjacent( self, node ):
        '''Return True iff `self` and `node` are adjacent.

        This is used to determine what nodes may be folded.
        '''
        return (self.parent==node or node.parent==self) \
        and self.group == node.group


    def iter_ancestors( self ):
        '''Iterate over each ancestor in decreasing rank order.'''
        parent = self.parent
        while parent:
            yield parent
            parent = parent.parent


    def is_ancestor_of( self, node ):
        '''Return True iff `node` is an ancestor of self.'''
        return node in self.iter_ancestors()


    # define comparisons such that parents come first

    def __lt__( self, other ):
        return other.is_ancestor_of( self )

    def __le__( self, other ):
        if self == other:
            return True
        return other.is_ancestor_of( self )

    def __gt__( self, other ):
        return self.is_ancestor_of( other )

    def __ge__( self, other ):
        if self == other:
            return True
        return self.is_ancestor_of( other )



class Graph( object ):
    no_group = False

    def __init__( self, initial_nodes=None ):
        self.nodes = Set( initial_nodes )


    def get_all_nodes( self ):
        '''Return Set of all nodes in graph.'''
        visited = Set()
        this_rank = Set(self.nodes)
        while this_rank:
            visited.union_update(this_rank)
            next_rank = Set()
            for node in this_rank:
                next_rank.union_update( node.all_referenced_nodes() )
            next_rank.difference_update(visited)
            this_rank = next_rank
        return visited


    def strip_merges_from_ancestry( self, merged=True, unmerged=False, parent=False, nodes=None, all_nodes=None ):
        '''Strip arcs from the set `nodes` and all their immediate ancestors.

        In a large graph it is useful to strip all merges from any mainline
        branches.  If `nodes` is unspecified, assume the set of root nodes in
        this Graph.

        the `merged`, `unmerged`, and `parent` parameters describe what types
        of arcs to strip. Default is to strip only merges.

        `all_nodes` should be the Set of all nodes from `get_all_nodes`, or
        None.
        '''

        if nodes is None:
            nodes = self.nodes
        if all_nodes is None:
            all_nodes = self.get_all_nodes()

        strip_from = nodes.copy()
        for node in nodes:
            strip_from.union_update( node.iter_ancestors() )

        for node in all_nodes:
            if merged and node.merged_nodes:
                node.merged_nodes.difference_update( strip_from )
            if unmerged and node.unmerged_nodes:
                node.unmerged_nodes.difference_update( strip_from )
            if parent and node not in strip_from:
                if node.parent in strip_from:
                    node.parent = None



    def fold_boring_nodes( self, threshold=2, all_nodes=None, foldable_nodes=None ):
        '''Fold boring sequences of nodes.
        
        Strings of nodes with identical sets of arcs at least `threshold` nodes
        long are removed, replaced with a node labeled "first..last".

        `all_nodes` should be the Set of all nodes from `get_all_nodes`, or
        None. The set of all nodes will be changed by this operation, so it
        must be regenerated with `get_all_nodes` if needed.

        `foldable_nodes` is a sequence of nodes to consider for folding. If
        None (default), all nodes are considered.
        '''

        if threshold < 2:
            raise ValueError( 'fold threshold may not be less than 2' )
        if all_nodes is None:
            all_nodes = self.get_all_nodes()
        if foldable_nodes is None:
            foldable_nodes = all_nodes

        # build backwards arcs
        merged_to_arcs = {}
        unmerged_to_arcs = {}
        tag_sources = Set()
        for node in all_nodes:
            for target in node.merged_nodes:
                merged_to_arcs.setdefault(target, []).append(node)
            for target in node.unmerged_nodes:
                unmerged_to_arcs.setdefault(target, []).append(node)
            if node.parent is not None \
            and node.group != node.parent.group:
                tag_sources.add(node.parent)

        foldable_sets = {}
        for node in foldable_nodes:
            if node.parent is not None and node.group != node.parent.group \
            or node in tag_sources:
                continue
            sig = (
                node.group,
                ImmutableSet( node.merged_nodes ),
                ImmutableSet( node.unmerged_nodes ),
                ImmutableSet( merged_to_arcs.get(node) ),
                ImmutableSet( unmerged_to_arcs.get(node) ))
            foldable_sets.setdefault(sig, []).append(node)

        del foldable_nodes
        del merged_to_arcs
        del unmerged_to_arcs
        del tag_sources


        def adjacent_spans( list ):
            split_at = []
            for i in xrange( len(list)-1 ):
                if not list[i].is_adjacent(list[i+1]):
                    split_at.append(i+1)
            if not split_at:
                yield list
            else:
                last_split = 0
                for split in split_at:
                    yield list[last_split:split]
                    last_split = split
                yield list[last_split:]


        renamed_nodes = {}
        for (group, merged, unmerged, merged_to, unmerged_to), foldables in foldable_sets.iteritems():
            if len(foldables) < threshold:
                continue
            foldables.sort()
            for span in adjacent_spans( foldables ):
                if len(span) < threshold:
                    continue
                fnode = Node('%s..%s' % (span[0], span[-1].label) )
                all_nodes.add(fnode)
                for renamed in span:
                    renamed_nodes[renamed] = fnode
                fnode.parent = span[0].parent
                fnode.group = span[0].group
                fnode.label = '%s..%s' % (span[0].label, span[-1].label)
                fnode.merged_nodes = span[0].merged_nodes
                fnode.unmerged_nodes = span[0].unmerged_nodes
                fnode.start_date = span[0].start_date
                fnode.end_date = span[-1].start_date
                if span[0].trunk:
                    fnode.trunk = True

        # update graph to reflect renamed nodes
        for node in all_nodes:
            try:
                node.parent = renamed_nodes[node.parent]
            except KeyError:
                pass

            if node.merged_nodes:
                renamed_here = node.merged_nodes.intersection( renamed_nodes )
                node.merged_nodes.difference_update( renamed_here )
                node.merged_nodes.union_update( [renamed_nodes[n] for n in renamed_here] )

            if node.unmerged_nodes:
                renamed_here = node.unmerged_nodes.intersection( renamed_nodes )
                node.unmerged_nodes.difference_update( renamed_here )
                node.unmerged_nodes.union_update( [renamed_nodes[n] for n in renamed_here] )


    def yield_group_roots(self, all_nodes=None):
        '''Yield the root nodes of each group.
        
        This is useful when adding full names to nodes that really need it.
        '''
        if all_nodes is None:
            all_nodes = self.get_all_nodes()
        for node in all_nodes:
            if node.parent is None \
            or node.parent.group != node.group:
                yield node


    def add_full_names(self, nodes):
        '''Prepend the group name to the labels of nodes.'''
        for node in nodes:
            if not node.group:
                continue
            # XXX we cheat here a bit and split the group name on /, which is
            # specific to tla.
            node.label = '%s\n%s' % (node.group.replace('/','\n'), node.label)


        
    def generate_output(self, options):
        '''Process the graph and generate output.'''
        raise NotImplementedError



class Stats( object ):

    def __init__( self, graph ):
        all_nodes = graph.get_all_nodes()
        self.node_count = len(all_nodes)

        edge_count = 0
        for node in all_nodes:
            edge_count += len(node.merged_nodes)
            edge_count += len(node.unmerged_nodes)
            if node.parent is not None:
                edge_count += 1
        self.edge_count = edge_count
        self.graph = graph

    def __str__( self ):
        return 'nodes: %i\tedges: %i ' % (self.node_count, self.edge_count)
